//
// DLL stuff
//
// by Denis Nechitailov, UU9JDR, 2002
//
// Based on modulator.c by Tomi Manninen, OH2BNS
// 

//------------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>

#include "modem.h"
#include "modemconfig.h"
#include "fec.h"
#include "filter.h"
#include "newqpsktx.h"
#include "complex.h"
#include "tbl.h"

#include "unistd.h"

#define SAMPLERATE(x)	((float)(x)*SymbolLen/DataCarriers/SymbolBits)

//------------------------------------------------------------------------------

typedef int (* TxPktProc)(void* procparam, unsigned char* pkt, unsigned len);

// Outer structure

struct TxState 
{
	TxPktProc pktproc; 
	void* procparam;

	struct txstate s; // have txstate inside
};

//------------------------------------------------------------------------------
//
// Allocate memory and init TxState structure
//

_declspec(dllexport) void *modconfig(unsigned int *samplerate, const char *params[])
{
	struct TxState *S; // = calloc(1, sizeof(struct txstate));
	struct txstate *s;

	if ((S = calloc(1, sizeof(struct TxState))) == NULL)
	{
		logprintf(MLOG_FATAL, "out of memory\n");
		return NULL;
	}

	s = &(S->s);

	s->chan = (struct modemchannel *)S;

	if (params[0]) {
		s->bps = strtoul(params[0], NULL, 0);
		if (s->bps < 1000)
			s->bps = 1000;
		if (s->bps > 5000)
			s->bps = 5000;
	} else
		s->bps = 2500;
	if (params[1]) {
		s->fec.inlv = strtoul(params[1], NULL, 0);
		if (s->fec.inlv < 0)
			s->fec.inlv = 0;
		if (s->fec.inlv > 16)
			s->fec.inlv = 16;
	} else
		s->fec.inlv = 8;
	if (params[2]) {
		s->fec.feclevel = strtoul(params[2], NULL, 0);
		if (s->fec.feclevel < 0)
			s->fec.feclevel = 0;
		if (s->fec.feclevel > 3)
			s->fec.feclevel = 3;
	} else
		s->fec.feclevel = 3;
	if (params[3]) {
		s->tunelen = strtoul(params[3], NULL, 0);
		if (s->tunelen < 0)
			s->tunelen = 0;
		if (s->tunelen > 64)
			s->tunelen = 64;
	} else
		s->tunelen = 32;
	if (params[4]) {
		s->synclen = strtoul(params[4], NULL, 0);
		if (s->synclen < 16)
			s->synclen = 16;
		if (s->synclen > 64)
			s->synclen = 64;
	} else
		s->synclen = 32;
	*samplerate = (int) (3.0 * SAMPLERATE(s->bps) + 0.5);
	return S;
}

//------------------------------------------------------------------------------
//
// Remove RxState structure
//

_declspec(dllexport) void modkill(void *state)
{
	struct TxState *S = (struct TxState *)state;
	// deinit???

	free(S);
}

//------------------------------------------------------------------------------
//
// Init the modulator
//

_declspec(dllexport) void modinit(void *state, unsigned int samplerate)
{
	struct TxState *S = (struct TxState *)state;
	struct txstate *s = &(S->s);
	float rate, f1, f2;

	rate = samplerate / SAMPLERATE(s->bps);
	s->bufsize = rate * SymbolLen + 16;
	f1 = 0.1;
	f2 = 0.9;
	init_tbl();
	init_filter(&s->filt, rate, f1, f2);
	init_fec(&s->fec);
	init_newqpsktx(s);
}

//------------------------------------------------------------------------------
//
// Fill the buffer
//
// return n (0<n<=numsamp) audio samples
// return 0 if nothing to fill and it's OK to go to RX
// 

_declspec(dllexport) int modmodulate(void *state, short* samples, int numsamp)
{
	struct TxState *S = (struct TxState *)state;
	struct txstate *s = &(S->s);
	complex *cbuf;
	int n, i;

	if (numsamp < s->bufsize)
	{
		logprintf(0, "Error: numsamp < s->bufsize\n");
		return 0; // Too bad, exit
	}

	cbuf = _alloca(s->bufsize * sizeof(complex));

	s->txdone = 0;

	n = newqpsktx(s, cbuf);
	for (i = 0; i < n; i++)
		samples[i] = (cbuf[i].re + cbuf[i].im) * 32768.0;

	if (s->txdone)
	{
		return 0;
	}

	return n;
}

//------------------------------------------------------------------------------
//
// Set modulator callback function
//

_declspec(dllexport) void modsetcallbacks(void *state, int (*pktproc)(void*, unsigned char*, unsigned), void* procparam)
{
	struct TxState *S = (struct TxState *)state;

	S->pktproc = pktproc;
	S->procparam = procparam;
}

//------------------------------------------------------------------------------
//
// Call the callback function to get the next byte to transmit
//

int pktget(struct modemchannel *chan, unsigned char *data, unsigned int len)
{
	struct TxState *S = (struct TxState *)chan;
    return (*S->pktproc)(S->procparam, data, len);
}

//------------------------------------------------------------------------------

